# coding=utf-8
"""
Contains the manager module for handling contract functionality within the events updater.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from typing import Dict, Any, Tuple

import brownie
from web3 import Web3
from web3.contract import Contract

from fastlane_bot.data.abi import BANCOR_V3_NETWORK_INFO_ABI, ERC20_ABI
from fastlane_bot.events.managers.base import BaseManager


class ContractsManager(BaseManager):
    def init_exchange_contracts(self):
        """
        Initialize the exchange contracts.
        """
        for exchange_name in self.SUPPORTED_EXCHANGES:
            self.event_contracts[exchange_name] = self.web3.eth.contract(
                abi=self.exchanges[exchange_name].get_abi(),
            )
            self.pool_contracts[exchange_name] = {}
            if exchange_name == "bancor_v3":
                self.pool_contracts[exchange_name][
                    self.cfg.BANCOR_V3_NETWORK_INFO_ADDRESS
                ] = brownie.Contract.from_abi(
                    address=self.cfg.BANCOR_V3_NETWORK_INFO_ADDRESS,
                    abi=BANCOR_V3_NETWORK_INFO_ABI,
                    name="BancorNetwork",
                )

    @staticmethod
    def get_or_create_token_contracts(
        web3: Web3, erc20_contracts: Dict[str, Contract], address: str
    ) -> Contract:
        """
        Get or create the token contracts.

        Parameters
        ----------
        web3 : Web3
            The Web3 instance.
        erc20_contracts : Dict[str, Contract]
            The ERC20 contracts.
        address : str
            The address.

        Returns
        -------
        Contract
            The token contract.

        """
        if address in erc20_contracts:
            contract = erc20_contracts[address]
        else:
            contract = web3.eth.contract(address=address, abi=ERC20_ABI)
            erc20_contracts[address] = contract
        return contract

    def add_pool_info_from_contract(
        self,
        exchange_name: str = None,
        address: str = None,
        event: Any = None,
    ) -> Dict[str, Any]:
        """
        Add the pool info from the contract.

        Parameters
        ----------
        exchange_name : str, optional
            The exchange name.
        address : str, optional
            The address.
        event : Any, optional
            The event.

        Returns
        -------
        Dict[str, Any]
            The pool info from the contract.

        """
        exchange_name = self.check_forked_exchange_names(exchange_name, address, event)
        if not exchange_name:
            self.cfg.logger.info(f"Exchange name not found {event}")
            return None

        if exchange_name not in self.SUPPORTED_EXCHANGES:
            self.cfg.logger.debug(
                f"Event exchange {exchange_name} not in exchanges={self.SUPPORTED_EXCHANGES} for address={address}"
            )
            return None

        pool_contract = self.get_pool_contract(exchange_name, address)
        self.pool_contracts[exchange_name][address] = pool_contract
        fee, fee_float = self.exchanges[exchange_name].get_fee(address, pool_contract)

        t0_addr = self.exchanges[exchange_name].get_tkn0(address, pool_contract, event)
        t1_addr = self.exchanges[exchange_name].get_tkn1(address, pool_contract, event)
        block_number = event["blockNumber"]

        return self.add_pool_info(
            address=address,
            exchange_name=exchange_name,
            fee=fee,
            fee_float=fee_float,
            tkn0_address=t0_addr,
            tkn1_address=t1_addr,
            cid=event["args"]["id"] if exchange_name == "carbon_v1" else None,
            contract=pool_contract,
            block_number=block_number,
        )

    def get_pool_contract(self, exchange_name: str, address: str) -> Contract:
        """
        Get the pool contract.

        Parameters
        ----------
        exchange_name : str
            The exchange name.
        address : str
            The address.

        Returns
        -------
        Contract
            The pool contract.

        """
        if exchange_name in self.exchanges:
            default_contract = self.web3.eth.contract(
                address=address, abi=self.exchanges[exchange_name].get_abi()
            )
            contract_key = (
                self.cfg.BANCOR_V3_NETWORK_INFO_ADDRESS
                if exchange_name == "bancor_v3"
                else address
            )
            return self.pool_contracts[exchange_name].get(
                contract_key, default_contract
            )
        else:
            return None

    def get_token_info_from_contract(
        self, web3: Web3, erc20_contracts: Dict[str, Contract], addr: str
    ) -> Tuple[str, int]:
        """
        Get the token info from contract.

        Parameters
        ----------
        web3 : Web3
            The web3 instance.
        erc20_contracts : Dict[str, Contract]
            The erc20 contracts.
        addr : str
            The address.

        Returns
        -------
        Tuple[str, int]
            The token info.

        """
        contract = self.get_or_create_token_contracts(web3, erc20_contracts, addr)
        try:
            return (
                contract.functions.symbol().call(),
                contract.functions.decimals().call(),
            )
        except Exception as e:
            self.cfg.logger.debug(f"Failed to get symbol and decimals for {addr} {e}")



# coding=utf-8
"""
Contains the pool class for Uniswap v2. This class is responsible for handling Uniswap v2 pools and updating the state of the pools.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from dataclasses import dataclass
from typing import Dict, Any

from web3.contract import Contract

from fastlane_bot.data.pools import uniswap_v2_pools
from fastlane_bot.events.pools.base import Pool


@dataclass
class UniswapV2Pool(Pool):
    """
    Class representing a Uniswap v2 pool.
    """

    exchange_name: str = "uniswap_v2"

    @staticmethod
    def unique_key() -> str:
        """
        see base class.
        """
        return "address"

    @classmethod
    def event_matches_format(cls, event: Dict[str, Any]) -> bool:
        """
        Check if an event matches the format of a Uniswap v2 event.
        """
        event_args = event["args"]
        return "reserve0" in event_args and event["address"] in uniswap_v2_pools

    def update_from_event(
        self, event_args: Dict[str, Any], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        See base class.
        """
        event_args = event_args["args"]
        data["tkn0_balance"] = event_args["reserve0"]
        data["tkn1_balance"] = event_args["reserve1"]
        for key, value in data.items():
            self.state[key] = value

        data["cid"] = self.state["cid"]
        data["fee"] = self.state["fee"]
        data["fee_float"] = self.state["fee_float"]
        data["exchange_name"] = self.state["exchange_name"]
        return data

    def update_from_contract(self, contract: Contract) -> Dict[str, Any]:
        """
        See base class.
        """
        reserve_balance = contract.caller.getReserves()
        params = {
            "fee": "0.003",
            "fee_float": 0.003,
            "tkn0_balance": reserve_balance[0],
            "tkn1_balance": reserve_balance[1],
            "exchange_name": "uniswap_v2",
        }
        for key, value in params.items():
            self.state[key] = value
        return params



"""
Testing utilities for the fastlane_bot package

USAGE

    from fastlane_bot.testing import *
"""
__VERSION__ = "1.2"
__DATE__ = "07/May/2023"

import math as m
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import os
import sys
from decimal import Decimal
import time
from itertools import zip_longest


def iseq(arg0, *args, eps=1e-6):
    """checks whether all arguments are equal to arg0, within tolerance eps if numeric"""
    if not args:
        raise ValueError("Must provide at least one arg", args)
    try:
        arg0+1
        isnumeric = True
    except:
        isnumeric = False
    #if isinstance(arg0, int) or isinstance(arg0, float):
    if isnumeric:
        # numeric testing
        if arg0 == 0:
            for arg in args:
                if abs(arg) > eps: 
                    return False
                return True
        for arg in args:
            if abs(arg/arg0-1) > eps:
                return False
            return True
    else:
        for arg in args:
            if not arg == arg0:
                return False
        return True

def raises(func, *args, **kwargs):
    """
    returns exception message if func(*args, **kwargs) raises, else False

    USAGE

        assert raises(func, 1, 3, three=3), "func(1, 2, three=3) should raise"
    """
    try:
        func(*args, **kwargs)
        return False
    except Exception as e:
        return str(e)


import re as _re

class VersionRequirementNotMetError(RuntimeError): pass

def _split_version_str(vstr):
    """splits version mumber string into tuple (int, int, int, ...)"""
    m = _re.match("^([0-9\.]*)", vstr.strip())
    if m is None:
        raise ValueError("Invalid version number string", vstr)
    vlst = tuple(int(x) for x in m.group(0).split("."))
    return vlst

def require(required, actual, raiseonfail=True, printmsg=True):
    """
    checks whether required version is smaller or equal actual version
    
    :required:      the required version, eg "1.2.1"
    :actual:        the actual version, eg "1.3-beta2"
    :raiseonfail:   if True, raises VersionRequirementNotMetError if requirements are not met
    :printmsg:      if True, prints message
    :returns:       True if requirements are met, False else*
    
    *note: "1.3-beta1" MEETS the requirement "1.3"!
    """
    rl, al = _split_version_str(required), _split_version_str(actual)
    #print(f"required={rl}, actual={al}")

    result = _require_version(rl,al)
    if printmsg:
        m1 = "" if result else "NOT "
        print(f"Version = {actual} [requirements >= {required} is {m1}met]")
    if not raiseonfail:
        return result
    if not result:
        raise VersionRequirementNotMetError(f"Version requirements not met (required = {rl}, actual = {al})", required, actual)

def _require_version(rl, al):
    """
    checks whether required version is smaller or equal actual version
    
    :rl:        the required version eg (1,2,1)
    :al:        the actual version, eg (1,3)
    :returns:   True if requirements are met, False else*
    """
    for r,a in zip_longest(rl, al, fillvalue=0):
        #print(f"r={r}, a={a}")
        if r > a:
            return False
        elif r < a:
            return True
    return True
    
    
print("imported m, np, pd, plt, os, sys, decimal; defined iseq, raises, require")



"""
Fastlane bot config -- logger
"""
__VERSION__ = "1.0"
__DATE__ = "03/May 2023"

from .base import ConfigBase
from . import selectors as S
import logging

# NOTE: THIS WHOLE LOGGING BUSINESS IS A BIT CONVOLUTED, SO AT ONE POINT
# WE MAY CONSIDER CLEANING IT UP

class ConfigLogger(ConfigBase):
    """
    Fastlane bot config -- logger
    
    :loglevel:    LOGLEVEL_DEBUG, LOGLEVEL_INFO (default), LOGLEVEL_WARNING, LOGLEVEL_ERROR
    """
    __VERSION__=__VERSION__
    __DATE__=__DATE__
    LOGGER_DEFAULT = S.LOGGER_DEFAULT
    LOGLEVEL_DEBUG = S.LOGLEVEL_DEBUG
    LOGLEVEL_INFO = S.LOGLEVEL_INFO
    LOGLEVEL_WARNING = S.LOGLEVEL_WARNING
    LOGLEVEL_ERROR = S.LOGLEVEL_ERROR
    LOGLEVEL = S.LOGLEVEL_INFO
    
    def get_logger(self, loglevel: str) -> logging.Logger:
        """
        Returns a logger with the specified logging level

        Args:
            loglevel (str): The desired logging level.

        Returns:
            logging.Logger: A logger object with the specified logging level.
        """
        log_level = getattr(logging, loglevel.upper())
        logger = logging.getLogger("fastlane")
        logger.setLevel(log_level)
        #print(f"[get_logger] {loglevel} {log_level}")
        handler = logging.StreamHandler()
        handler.setLevel(log_level)
        formatter = logging.Formatter(
            "%(asctime)s [%(name)s:%(levelname)s] - %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger
    
    @property
    def logger(self):
        return self._logger
    
    def debug(self, *args, **kwargs):
        """calls logger.debug()"""
        return self.logger.debug(*args, **kwargs)
    
    def info(self, *args, **kwargs):
        """calls logger.info()"""
        return self.logger.info(*args, **kwargs)
    
    def warning(self, *args, **kwargs):
        """calls logger.warning()"""
        return self.logger.warning(*args, **kwargs)
    
    def error(self, *args, **kwargs):
        """calls logger.error()"""
        return self.logger.error(*args, **kwargs)
    
    @classmethod
    def new(cls, *, logger=None, loglevel=None, **kwargs):
        """
        Return a new ConfigLogger.
        """
        if logger is None:
            logger = S.LOGGER_DEFAULT
        
        if logger == S.LOGGER_DEFAULT:
            return _ConfigLoggerDefault(_direct=False, loglevel=loglevel, **kwargs)
        else:
            raise ValueError(f"Unknown logger: {logger}")
        
    def __init__(self, loglevel=None, **kwargs):
        super().__init__(**kwargs)
        #print("[ConfigLogger]", loglevel, self.LOGLEVEL)
        if not loglevel is None:
            assert loglevel in {
                self.LOGLEVEL_DEBUG, 
                self.LOGLEVEL_INFO, 
                self.LOGLEVEL_WARNING, 
                self.LOGLEVEL_ERROR
            }, f"unknown loglevel {loglevel}"
            self.LOGLEVEL = loglevel
        self._logger = self.get_logger(self.LOGLEVEL)


class _ConfigLoggerDefault(ConfigLogger):
    """
    Fastlane bot config -- logger
    """
    pass
        

    



# coding=utf-8
"""
Contains the factory class for exchanges. This class is responsible for creating exchanges.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""


class ExchangeFactory:
    """
    Factory class for exchanges
    """

    def __init__(self):
        self._creators = {}

    def register_exchange(self, key, creator):
        """
        Register an exchange with the factory

        Parameters
        ----------
        key : str
            The key to use for the exchange
        creator : Exchange
            The exchange class to register
        """
        self._creators[key] = creator

    def get_exchange(self, key):
        """
        Get an exchange from the factory

        Parameters
        ----------
        key : str
            The key to use for the exchange

        Returns
        -------
        Exchange
            The exchange class
        """
        creator = self._creators.get(key)
        if not creator:
            raise ValueError(key)
        return creator()



# coding=utf-8
"""
This module contains the tests for the exchanges classes
"""
import json

from fastlane_bot import Bot
from fastlane_bot.tools.cpc import ConstantProductCurve as CPC
from fastlane_bot.events.exchanges import (
    UniswapV2,
    UniswapV3,
    SushiswapV2,
    CarbonV1,
    BancorV3,
)
from fastlane_bot.data.abi import (
    UNISWAP_V2_POOL_ABI,
    UNISWAP_V3_POOL_ABI,
    SUSHISWAP_POOLS_ABI,
    BANCOR_V3_POOL_COLLECTION_ABI,
    CARBON_CONTROLLER_ABI,
)
from unittest.mock import Mock
import pytest

print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CPC))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(Bot))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV2))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(UniswapV3))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(SushiswapV2))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonV1))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(BancorV3))
from fastlane_bot.testing import *

plt.style.use("seaborn-dark")
plt.rcParams["figure.figsize"] = [12, 6]
from fastlane_bot import __VERSION__

require("3.0", __VERSION__)


@pytest.fixture
def setup_data():
    with open("fastlane_bot/data/event_test_data.json", "r") as f:
        setup_data = json.load(f)
    return setup_data


# The contract fixture creates a mocked contract with predefined return values
@pytest.fixture
def mocked_contract():
    mocked_contract = Mock()
    mocked_contract.functions.token0.return_value.call.return_value = "token0"
    mocked_contract.functions.token1.return_value.call.return_value = "token1"
    mocked_contract.functions.fee.return_value.call.return_value = 3000
    return mocked_contract


def test_uniswap_v2_exchange(setup_data, mocked_contract):
    uniswap_v2_exchange = UniswapV2()
    assert (
        uniswap_v2_exchange.get_abi() == UNISWAP_V2_POOL_ABI
    )  # replace with your constant
    assert uniswap_v2_exchange.get_fee("", mocked_contract) == ("0.003", 0.003)
    assert (
        uniswap_v2_exchange.get_tkn0("", mocked_contract, {})
        == mocked_contract.functions.token0().call()
    )
    assert (
        uniswap_v2_exchange.get_tkn1("", mocked_contract, {})
        == mocked_contract.functions.token1().call()
    )


def test_uniswap_v3_exchange(setup_data, mocked_contract):
    uniswap_v3_exchange = UniswapV3()
    assert (
        uniswap_v3_exchange.get_abi() == UNISWAP_V3_POOL_ABI
    )  # replace with your constant
    assert uniswap_v3_exchange.get_fee("", mocked_contract) == (
        mocked_contract.functions.fee().call(),
        float(mocked_contract.functions.fee().call()) / 1e6,
    )
    assert (
        uniswap_v3_exchange.get_tkn0("", mocked_contract, {})
        == mocked_contract.functions.token0().call()
    )
    assert (
        uniswap_v3_exchange.get_tkn1("", mocked_contract, {})
        == mocked_contract.functions.token1().call()
    )


def test_sushiswap_v2_exchange(setup_data, mocked_contract):
    sushiswap_v2_exchange = SushiswapV2()
    assert (
        sushiswap_v2_exchange.get_abi() == SUSHISWAP_POOLS_ABI
    )  # replace with your constant
    assert sushiswap_v2_exchange.get_fee("", mocked_contract) == ("0.003", 0.003)
    assert (
        sushiswap_v2_exchange.get_tkn0("", mocked_contract, {})
        == mocked_contract.functions.token0().call()
    )
    assert (
        sushiswap_v2_exchange.get_tkn1("", mocked_contract, {})
        == mocked_contract.functions.token1().call()
    )


def test_bancor_v3_exchange(setup_data, mocked_contract):
    bancor_v3_exchange = BancorV3()
    assert (
        bancor_v3_exchange.get_abi() == BANCOR_V3_POOL_COLLECTION_ABI
    )  # replace with your constant
    assert bancor_v3_exchange.get_fee("", mocked_contract) == ("0.000", 0.000)
    assert (
        bancor_v3_exchange.get_tkn0("", mocked_contract, setup_data["bancor_v3_event"])
        == bancor_v3_exchange.BNT_ADDRESS
    )
    assert (
        bancor_v3_exchange.get_tkn1("", mocked_contract, setup_data["bancor_v3_event"])
        == setup_data["bancor_v3_event"]["args"]["pool"]
        if setup_data["bancor_v3_event"]["args"]["pool"]
        != bancor_v3_exchange.BNT_ADDRESS
        else setup_data["bancor_v3_event"]["args"]["tkn_address"]
    )


def test_carbon_v1_exchange_update(setup_data, mocked_contract):
    carbon_v1_exchange = CarbonV1()
    assert (
        carbon_v1_exchange.get_abi() == CARBON_CONTROLLER_ABI
    )  # replace with your constant
    assert carbon_v1_exchange.get_fee("", mocked_contract) == ("0.002", 0.002)
    assert (
        carbon_v1_exchange.get_tkn0(
            "", mocked_contract, setup_data["carbon_v1_event_update"]
        )
        == setup_data["carbon_v1_event_update"]["args"]["token0"]
    )
    assert (
        carbon_v1_exchange.get_tkn1(
            "", mocked_contract, setup_data["carbon_v1_event_update"]
        )
        == setup_data["carbon_v1_event_update"]["args"]["token1"]
    )


def test_carbon_v1_exchange_create(setup_data, mocked_contract):
    carbon_v1_exchange = CarbonV1()
    assert (
        carbon_v1_exchange.get_abi() == CARBON_CONTROLLER_ABI
    )  # replace with your constant
    assert carbon_v1_exchange.get_fee("", mocked_contract) == ("0.002", 0.002)
    assert (
        carbon_v1_exchange.get_tkn0(
            "", mocked_contract, setup_data["carbon_v1_event_create"]
        )
        == setup_data["carbon_v1_event_create"]["args"]["token0"]
    )
    assert (
        carbon_v1_exchange.get_tkn1(
            "", mocked_contract, setup_data["carbon_v1_event_create"]
        )
        == setup_data["carbon_v1_event_create"]["args"]["token1"]
    )


def test_carbon_v1_exchange_delete(setup_data, mocked_contract):
    carbon_v1_exchange = CarbonV1()
    assert carbon_v1_exchange.get_abi() == CARBON_CONTROLLER_ABI
    cid = setup_data["carbon_v1_event_delete"]["args"]["id"]
    assert cid not in [strat["cid"] for strat in carbon_v1_exchange.pools]


#%%



# coding=utf-8
"""
Carbon Arbitrage Bot setup.py installer.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""

import re

from setuptools import find_packages, setup

import os
from dotenv import load_dotenv

# Load .env file
env_file = '.env'
try:
    assert os.path.exists(env_file), "The .env file is missing. See README.md for instructions"
except AssertionError as e:
    # Create .env file
    print("Creating .env file. Please update it with your environment variables. See README.md for instructions")
    with open(env_file, 'w') as f:
        f.write(f"export WEB3_ALCHEMY_PROJECT_ID=\n")
        f.write(f"export ETH_PRIVATE_KEY_BE_CAREFUL=\n")
        f.write(f"export DEFAULT_MIN_PROFIT_BNT=200\n")
        f.write(f"export ETHERSCAN_TOKEN=\n")
        f.write(f"export TENDERLY_FORK_ID=\n")
        f.write(f"export TENDERLY_ACCESS_KEY=\n")
        f.write(f"export TENDERLY_PROJECT=\n")
        f.write(f"export TENDERLY_USER=\n")

load_dotenv(env_file)

# Check for required variables
required_vars = ['WEB3_ALCHEMY_PROJECT_ID',
                 'ETH_PRIVATE_KEY_BE_CAREFUL',
                 'DEFAULT_MIN_PROFIT_BNT',
                 'ETHERSCAN_TOKEN',
                 'TENDERLY_FORK_ID',
                 'TENDERLY_ACCESS_KEY',
                 'TENDERLY_PROJECT',
                 'TENDERLY_USER']

with open(env_file, 'a') as f:
    for var in required_vars:
        if var not in os.environ:
            print(f"The {var} environment variable is missing in .env file. Adding it with a default empty value. See README.md for instructions")
            if var == 'DEFAULT_MIN_PROFIT_BNT':
                f.write(f"export {var}=1\n")
                os.environ[var] = '1'
            else:
                f.write(f"export {var}=\n")
                os.environ[var] = ''  # Optionally update the current environment as well
        if os.environ[var] == '' and 'TENDERLY' not in var and 'ETHERSCAN_TOKEN' not in var:
            raise Exception(f"The {var} environment variable cannot be None. Please update the .env file. See README.md for instructions")

import brownie_setup

with open("fastlane_bot/__init__.py") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    )[1]

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

extras_require = {}
extras_require["complete"] = sorted(set(sum(extras_require.values(), [])))

setup(
    name="fastlane_bot",
    version=version,
    author="Bancor Network",
    author_email="mike@bancor.network",
    description="""
                    Carbon Arbitrage Bot, an open-source arbitrage protocol, allows any user to perform arbitrage between Bancor ecosystem protocols and external exchanges and redirect arbitrage profits back to the protocol.
                """,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bancorprotocol/carbon-bot",
    install_requires=open("requirements.txt").readlines(),
    extras_require=extras_require,
    tests_require=["pytest~=6.2.5", "pytest-mock~=3.10.0", "imgkit~=1.2.3"],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">= 3.8, != 3.11.*",
)
